#ifndef __MDFN_NES_INPUT_SHARE_H
#define __MDFN_NES_INPUT_SHARE_H

#include "../nes.h"
#include "../input.h"
#include "../ppu/ppu.h"
#include "../x6502.h"
#include "../ppu/palette.h"

#include "cursor.h"

using namespace MDFN_IEN_NES;

#endif
